/*
* Copyright (C) Mellanox Technologies Ltd. 2018.  ALL RIGHTS RESERVED.
*
* See file LICENSE for terms.
*/
#ifndef API_CLX_API_H_
#define API_CLX_API_H_

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#ifdef __cplusplus
# define BEGIN_C_DECLS  extern "C" {
# define END_C_DECLS    }
#else
# define BEGIN_C_DECLS
# define END_C_DECLS
#endif

typedef void *   clx_api_schema_t;
typedef uint64_t clx_api_timestamp_t;
typedef uint8_t  clx_api_type_index_t;


typedef union clx_api_version_t {
    struct {
        uint8_t    major;
        uint8_t    minor;
        uint8_t    patch;
        uint8_t    date;
    };
    uint32_t hex;
} clx_api_version_t;


/** @struct clx_api_provider_t
 * @brief Provider that is used by clx_api countext to
 */
typedef struct clx_api_provider_t {
    clx_api_version_t   version;       ///< Version of provider. Will be saved as schema version.
    char*               name;          ///< Provider or counters group identifier.
    char*               description;   ///< Detailed description of the provider.
    void*               data;          ///< Implementation-specific provider data

    bool (*initialize)(void*ctx,  struct clx_api_provider_t* provider);  ///< Initializeer callback, should be implemented by user
                                                                         ///< Add counters / event types in this callback.
    void (*finalize)(void* ctx, struct clx_api_provider_t* provider);
} clx_api_provider_t;


#define CLX_API_DATA_PATH_TEMPLATE          "{{year}}/{{month}}{{day}}/{{source}}/{{tag}}{{id}}.bin"


/** @struct clx_api_params_t
 *  @brief Parameters for clx_api containing configuration of data destination.
 *
 * Data will be collected into api buffer and will be written to binary files according to data_path_template or
 * will be exported if ipc/fluent bit/prometheus are enabled.
 * The recommended default for data_path_template is: CLX_DATA_PATH_TEMPLATE="{{year}}/{{month}}{{day}}/{{source}}/{{tag}}{{id}}.bin"
 * Using the convention garantees not having unmanagable number of files in a single directory
 *
 * Data source_id and source_tag will be used as a part of data_path_template for writing data.
 */
typedef struct clx_api_params_t {
    char *source_id;           /**< {source} in data_path_template. Set to node name, guid, etc */
    char *source_tag;          /**< {tag} in data_path_template. Use empty-string tag for counters and non-empty for events. */

    uint64_t buffer_size;      /**< Size of the internal buffer in bytes. Set to 0 to use default 60 * 1024 or set manually */

    // File writer parameters.
    bool  file_write_enabled;  /**< set to 1 to enable, to 0 to disable. */
    char* data_root;           /**< data_root in which data will be stored according to the naming convention*/
    char* schema_path;         /**< path to schema folder. Schema contains the meta-data which allows access binary data. */

    char* data_path_template;  /**< Defines the naming convention to be used when writing files */

    /// Once one of the limits is reached, the current file is closed and new a file is generated.
    size_t max_file_size;      /**< File size limit in bytes*/
    clx_api_timestamp_t max_file_age; /**< Time limit for using current binary file in microseconds*/

    // IPC parameters
    bool  ipc_enabled;          /**< Flag for enabling IPC. set to 1 to enable IPC or to 0 to disable. */
    char* ipc_sockets_dir;      /**< Path to ipc sockets dir. Should contain clx ipc socket which is set in clx_config.ini. Recommended default is "/tmp/ipc_sockets". */

    // TBD(romanpr): create parameters for prometheus and fluent bit. Now exporters can be enabled only via env variables
    // uint16_t prometheus_port;
    // char* prometheus_host;
    char* prometheus_enpoint;
    char* cset_dir_path;
} clx_api_params_t;



/** @struct clx_api_value_type_t
 * @brief clx_api_value_type_t is a mode that will be accounted while querying data with clxcli.
 */
typedef enum clx_api_value_type_t {
    CLX_FIELD_VALUE_ABSOLUTE,    ///< Data will be queried as it was collected. Reccomended to use with counters that can increase and decrease.
    CLX_FIELD_VALUE_RELATIVE,    ///< Data will be normalized according to first value. Can be used for non-decreasing counters.
    CLX_FIELD_VALUE_DERIVATIVE,  ///< Queried data will be displayed as differences between adjacent values.
} clx_api_value_type_t;

/** @struct clx_api_data_type_t
 * @brief List of data types for counters.
 *
 */
typedef enum clx_api_data_type_t {
    CLX_API_DATA_TYPE_UNDEFINED = 0,

    // Scalar types
    CLX_API_DATA_TYPE_UINT64,
    CLX_API_DATA_TYPE_FP64,
    CLX_API_DATA_TYPE_BIT64,
    CLX_API_DATA_TYPE_INT64,
    CLX_API_DATA_TYPE_STRING,
} clx_api_data_type_t;


/** @struct clx_api_counter_info_t
 * @brief structure that describes counter meta information.
 *
 */
typedef struct clx_api_counter_info_t {
    char*                  counter_name;   ///< Name for querying data
    char*                  description;    ///< Detailed description of counter.
    char*                  units;          ///< String of type "Mb/sec", "Kb/sec", etc. Set to empty string if not used.
    clx_api_data_type_t    value_type;     ///< Collectx data type for counter value.
    clx_api_value_type_t   counting_mode;  ///< Mode for querying collected data by clxcli.
    uint32_t               offset;         ///< Offset for counter value in counters buffer. Computed automatically by clx_api.
    uint32_t               length;         ///< Lenfth of counter in bytes. Set for string type.
} clx_api_counter_info_t;


/** @struct clx_api_event_field_info_t
 * @brief clx_api_event_field_info_t is a structure that describes event field meta information.
 *
 * Event types are based on CollectX built-in types. Consider the following table:
 *   |  type name           |   CollectX alias                     |
 *   |----------------------|--------------------------------------|
 *   |  bool                |   BOOL                               |
 *   |  char                |   CHAR                               |
 *   |  short               |   SHORT                              |
 *   |  int                 |   INT                                |
 *   |  long                |   LONG                               |
 *   |  long long           |   LONGLONG                           |
 *   |  unsigned char       |   UCHAR                              |
 *   |  unsigned short      |   USHORT                             |
 *   |  unsigned int        |   UINT                               |
 *   |  unsigned long       |   ULONG                              |
 *   |  unsigned long long  |   ULONGLONG                          |
 *   |  float               |   FLOAT                              |
 *   |  double              |   DOUBLE                             |
 *   |  int8_t              |   INT8                               |
 *   |  int16_t             |   INT16                              |
 *   |  int32_t             |   INT32                              |
 *   |  int64_t             |   INT64                              |
 *   |  uint8_t             |   UINT8                              |
 *   |  uint16_t            |   UINT16                             |
 *   |  uint32_t            |   UINT32                             |
 *   |  uint64_t            |   UINT64                             |
 *   |  uint64_t            |   CLX_TYPE_TIMESTAMP or "timestamp"  |
 *   |  NULL                |   CLX_TYPE_NONE                      |
 */
typedef struct clx_api_event_field_info_t {
    const char*          field_name;       ///< Name of event field
    const char*          description;      ///< Event field descripion
    const char*          type_name;        ///< Name of Collectx built in type.
    clx_api_value_type_t counting_mode;    ///< Mode for querying collected data by clxcli.
    uint16_t             array_length;     ///< Array length for this event type. If set to 1 will be used as single value.
} clx_api_event_field_info_t;

#define NUM_OF_FIELDS(type) (sizeof(type)/sizeof(clx_api_event_field_info_t))

typedef enum clx_api_ts_error_t {
    CLX_API_OK = 0,
    CLX_API_NOMEMORY,
    CLX_API_DUPLICATE,
    CLX_API_UNDEFINED,
    CLX_API_NOT_IMPLEMENTED,
} clx_api_ts_error_t;


#ifndef HAVE_EVENT_CONTEXT
typedef void clx_api_context_t;
#endif


BEGIN_C_DECLS
// ======================= INITIALIZE / DESTOROY FUNCTIONS =====================
/**
 * @brief Creates and returns API context based on parameters and provider
 *
 * @param p        Parameters with the configured data destinations.
 * @param provider Defines data to be collected. Implement .initialize callback with creation of counters / events.
 * @return         Pointer to clx_api context.
 */
void* clx_api_create_context(clx_api_params_t* p, clx_api_provider_t* provider);

/**
 * @brief Clones and returns initial context vctx according to new parameters.
 *
 * Cloned contexts use the same provider, e.i. it will share the same schema.
 * Use clones when the same data should be collected from different sources.
 * Each clone will write the data to its own file, while all clones will share the same exporters.
 *
 * @param vctx Pointer to clx_api context.
 * @param p    parameters of original clx_api context with new source_id / source_tag.
 * @return     Pointer to clx_api context clone.
 * @return     NULL on fail.
 */
void* clx_api_clone_context(void* vctx, clx_api_params_t* p);

/**
 * @brief Destroys clx_api context. Should be applied to all clones and original context.
 *
 * @param vctx Pointer to clx_api context.
 * @return     true
 * @return     false
 */
bool clx_api_destroy_context(void* vctx);


// ========================= CONFIGURATION FUNCTIONS ===========================
clx_api_params_t* clx_api_get_params(void* vctx);
/// Get the parameters of context / clone `vctx`.

/**
 * @brief Check if file write is enabled for the clx_api context / clone.
 *
 * @param vctx Pointer to clx_api context
 * @return     true
 * @return     false
 */
bool clx_api_is_file_write_enabled(void* vctx);


/**
 * @brief Get binary file timestamp for the clx_api context / clone.
 *
 * @param vctx  Pointer to clx_api context
 * @return      timestamp of the current bin file.
 */
uint64_t clx_api_get_file_timestamp(void* vctx);


/**
 * @brief Get the counters schema size for the clx_api context / clone.
 *
 * @param vctx  Pointer to clx_api context
 * @return      counters schema size of the context.
 */
uint64_t clx_api_get_counters_schema_size(void* vctx);


/**
 * @brief Get current data root of the clx_api context / clone.
 *
 * @param vctx Pointer to clx_api context.
 * @return Path to data root.
 */
char* clx_api_get_data_root(void* vctx);


// ============================ DATA FUNCTIONS =================================
void clx_api_force_write(void* vctx);

/**
 * @brief Creates event type and stores it into clx_api schema.
 *
 * Fills the index of the event in the schema. Call this only in the provider's .initialize-callback.
 *
 * @param vctx       Pointer to clx_api context.
 * @param event_name Type name which will be used in shcema.
 * @param fields     List of event fields.
 * @param num_fields Number of event fields
 * @param index      This field will be filled with the event type index in the schema.
 * @return     true
 * @return     false
 */
bool clx_api_add_event_type(void *vctx, char *event_name, clx_api_event_field_info_t* fields, int num_fields, uint8_t* index);


/**
 * @brief Creates event type and stores it into clx_api schema.
 *
 * Fills the index of the event in the schema. Call this only in the provider's .initialize-callback.
 *
 * @param vctx       Pointer to clx_api context.
 * @param event_data Buffer with consecutive events of the same type.
 * @param ti         Event type in to write.
 * @param num_events Number of events in event_data buffer.
 * @return     true
 * @return     false
 */
bool clx_api_write_event(void* vctx, void* event_data, clx_api_type_index_t ti, int num_events);


/**
 * @brief Creates new counter and adds it to schema group.
 *
 * Fills the index of the counter in the schema group. Call this only in the provider's .initialize-callback.
 *
 * @param vctx        Pointer to clx_api context.
 * @param counter     Counter info to add.
 * @param group_name  Name of schema counter group where counter will be placed.
 * @param counter_num This field will be filled with the counter index in the group.
 * @return     true
 * @return     false
 */
bool clx_api_add_counter(void* vctx, clx_api_counter_info_t* counter, char* group_name, uint32_t* counter_num);


/**
 * @brief Get counters buffer to write counters manually.
 *
 * Counters buffer will be created as a free memory segment of clx_api internal buffer.
 * If there is not enough space, it will flush previously written data. Returns NULL on error.
 *
 * @param vctx        Pointer to clx_api context.
 * @param counter     Counter info to add.
 * @param group_name  Name of schema counter group where counter will be placed.
 * @param counter_num This field will be filled with the counter index in the group.
 * @return            Pointer to the clx_api buffer.
 * @return            NULL on the failure.
 */
void* clx_api_get_counters_buffer(void* vctx, uint64_t timestamp, uint32_t *data_size);


// =============================== IPC FUNCTIONS ===============================
void clx_api_ipc_connect_if_needed(void* vctx);
/// Use to make sure that IPC connected. If clx process restarted for some reason it will reestablish the connection.


// ================================== UTILS ====================================
/**
 * @brief Get the current time in CollectX format.
 *
 * @return Current timestamp in microsecinds.
 */
uint64_t clx_api_get_timestamp(void);


/**
 * @brief Set the first timestamp for current data_page.
 *
 * Use to initialize first binary file start timestamp to non-default value if needed.
 * If the app that uses clx_api will restart, it will write to the new file.
 * To continue to write to the previous file reset timestamp according to the timestamp in the previous file.
 *
 * @param vctx      Pointer to clx_api context.
 * @param timestamp Timestamp to set.
 */
void clx_api_set_first_timestamp(void* vctx, uint64_t timestamp);


// ============================================================================ //
// =========================      CLX API READ      =========================== //
// ============================================================================ //


// ========================= INITIALIZATION FUNCTIONS ==========================

/**
 * @brief Creates and returns application context.
 *
 * Reads all counter schemas from directory "schema_dir".
 * @param schema_dir
 * @return void* pointer to created context.
 * @return NULL on error
 */
void* clx_api_read_create_context(char* schema_dir);


/**
 * @brief  Read and reatunt counter binary file meta information.
 *
 * Opens counter binary file \p filename, and returns file meta info. If schema for that file
 * was not found in \p schema_dir previously, tries to find schema from the binary
 * file folder.
 *
 * @param context  ointer to CLX API READ context
 * @param filename absolute or relative path to counters binary file.
 * @return void* pointer to file meta data.
 * @return NULL on error
 */
void* clx_api_open_counter_file(void* context, char* filename);


/**
 * @brief Creates and returns empty counter set for current \p file meta info.
 *
 * @param context pointer to CLX API READ context
 * @param file    pointer to file meta info
 * @return void* pointer to an empty counterset.
 * @return NULL on error
 */
void* clx_api_create_counterset(void* context, void* file);


// ============================ DATA READ FUNCTIONS ============================
/**
 * @brief Allocates and returns pointer to a buffer for the counterset.
 *
 * Buffer size depends on number and types of counters included in cset.
 * @param cset pointer to counterset.
 * @return void* pointer to buffer.
 * @return NULL on error
 */
void* clx_api_allocate_counters_buffer(void* cset);


/**
 * @brief Reads single block of counters data from \p file into the \p data buffer.
 * Only counters from \p cset are written to \p data buffer. The timestamp of the
 * data block will be written to \p timestamp.
 *
 * @param file       pointer to counters binary file meta information.
 * @param cset       pointer to counterset.
 * @param[out] timestamp  timestamp of the data block in usec.
 * @param source
 * @param data       counters data buffer that fits \p cset.
 * @return 1 if no more data to read OR cannot read the next data block
 * @return 0 if not finished.
 */
int clx_api_get_next_data(void* file, void* cset, uint64_t *timestamp, char* source, void* data);


// ========================== COUNTERS INFO FUNCTIONS ==========================
/**
 * @brief Returns all the counters of schema corresponding to input \p file and updates \p num_counters.
 *
 * @param context pointer to CLX API READ context
 * @param file    pointer to counters binary file meta information.
 * @param[out] num_counters number of all available counters.
 * @return clx_api_counter_info_t* array of all available counters.
 */
clx_api_counter_info_t* clx_api_get_all_counters(void* context, void* file, int *num_counters);


/**
 * @brief Returns all the counters included to the counter set \p cset.
 *
 * @param cset pointer to counterset.
 * @return clx_api_counter_info_t* array of counters from \p cset.
 */
clx_api_counter_info_t* clx_api_get_counters(void* cset);


/**
 * @brief Returns number of counters contained by counter set \p cset.
 *
 * @param cset pointer to counterset.
 * @return int number of counters from \p cset.
 */
int clx_api_get_num_counters(void* cset);


// ========================== ADD COUNTERS FUNCTIONS ===========================
/**
 * @brief Adds a counter matching exact \p name to counter set \p cset, if it is not already included.
 *
 * @param cset pointer to counterset.
 * @param name exact counter name.
 * @return  1 counter already present
 * @return  0 counter was added
 * @return -1 if counter not found
 */
int clx_api_add_counter_exact(void* cset, char* name);


/**
 * @brief Adds all the counters matching token \p tok to counter set \p cset, that were not included previously.
 *
 * \p tok is a name fragment or a string of format "tok1+tok2-tok".
 *In the last example, considers only counter names that
 * include both "tok1" and "tok2", and not include "tok3".
 * To match counter name from the begining use "^tok", for name end use "tok$",
 * otherwise it will look for matches everywhere.
 * @param cset pointer to counterset.
 * @param tok a name fragment or a string of several fragments separated with + or -.
 * @return 1 on success
 * @return -1 on failure
 */
int clx_api_add_counters_matching(void* cset, char* tok);


/**
 * @brief Adds all available counters to counter set \p cset.
 *
 * @param cset pointer to counterset.
 * @return 1 on success
 * @return -1 on failure
 */
int clx_api_add_all_counters(void* cset);


// =========================== DATA ACCESS FUNCTIONS ===========================
/**
 * @brief Returns type of the \p idx -th counter of the counter set \p cset.
 * Available types are:
 *                      CLX_API_DATA_TYPE_UINT64,
 *                      CLX_API_DATA_TYPE_FP64,
 *                      CLX_API_DATA_TYPE_BIT64,
 *                      CLX_API_DATA_TYPE_INT64.
 * @param cset pointer to counter set.
 * @param idx index of counter in the \p cset.
 * @return clx_api_data_type_t counter value type.
 */
clx_api_data_type_t clx_api_get_type(void* cset, unsigned idx);


/**
 * @brief Read double value of \p idx counter from counter set \p cset from buffer \p data.
 *
 * @param cset pointer to counter set.
 * @param idx  index of counter in the \p cset.
 * @param data data buffer that is consistent with \p cset.
 * @return double counter value
 */
double   clx_api_get_double(void* cset, unsigned idx, void* data);


/**
 * @brief Read int64_t value of \p idx counter from counter set \p cset from buffer \p data.
 *
 * @param cset pointer to counter set.
 * @param idx  index of counter in the \p cset.
 * @param data data buffer that is consistent with \p cset.
 * @return int64_t counter value
 */
int64_t  clx_api_get_int64(void* cset, unsigned idx, void* data);


/**
 * @brief Read uint64_t value of \p idx counter from counter set \p cset from buffer \p data.
 *
 * @param cset pointer to counter set.
 * @param idx  index of counter in the \p cset.
 * @param data data buffer that is consistent with \p cset.
 * @return uint64_t counter value
 */
uint64_t clx_api_get_uint64(void* cset, unsigned idx, void* data);


/**
 * @brief Read string value of \p idx counter from counter set \p cset from buffer \p data.
 *
 * @param cset pointer to counter set.
 * @param idx  index of counter in the \p cset.
 * @param data data buffer that is consistent with \p cset.
 * @return string counter value
 */
char*    clx_api_get_str(void* cset, unsigned idx, void* data);


// ============================= CLEANUP FUNCTIONS =============================
/**
 * @brief Application context destructor.
 *
 * @param context pointer to CLX API READ context
 */
void clx_api_read_destroy_context(void* context);


/**
 * @brief Counter set destructor. Apply on each counter set.
 *
 * @param cset pointer to counter set.
 */
void clx_api_destroy_counter_set(void* cset);
//

/**
 * @brief File meta info destructor. Apply on each file.
 *
 * @param file    pointer to counters binary file meta information.
 */
void clx_api_destroy_and_close_file(void* file);



END_C_DECLS

#endif  // API_CLX_API_H_
